              (* +----------------------------------+
                 |  Boite  outils sur les taquins  |
                 +----------------------------------+ *)

type position = {
  table : int vect vect; (* matrice des pices *)
  i0 : int;              (* coordonnes de la case vide *)
  j0 : int
};;

type mouvement = Nord | Sud | Est | Ouest;;

(* inversion des mouvements *)
let inverse = function
  | Nord -> Sud   | Sud   -> Nord
  | Est  -> Ouest | Ouest -> Est
;;

(* position  obtenir *)
let but n p =
  let m = make_matrix n p 0 in
  for i=0 to n-1 do for j=0 to p-1 do
    m.(i).(j) <- p*i + j
  done done;
  {table = m; i0 = 0; j0 = 0}
;;

(* liste des mouvements possibles *)
let possibles t =
  let n = vect_length t.table
  and p = vect_length t.table.(0) in
  (if t.i0 > 0   then [Nord]  else []) @
  (if t.i0 < n-1 then [Sud]   else []) @
  (if t.j0 < p-1 then [Est]   else []) @
  (if t.j0 > 0   then [Ouest] else [])
;;

(* effectue un mouvement *)
let dplace t m = 
  let n = vect_length t.table
  and p = vect_length t.table.(0) in
  let i1,j1 = match m with
  | Nord  -> (t.i0-1, t.j0)
  | Sud   -> (t.i0+1, t.j0)
  | Est   -> (t.i0,   t.j0+1)
  | Ouest -> (t.i0,   t.j0-1)
  in
  let table' = make_vect n [||] in
  for i=0 to n-1 do table'.(i) <- copy_vect t.table.(i) done;
  table'.(i1).(j1)     <- t.table.(t.i0).(t.j0);
  table'.(t.i0).(t.j0) <- t.table.(i1).(j1);
  {table = table'; i0=i1; j0=j1}
;;

(* taquin rsoluble en au plus k mouvements *)
let random_position n p k =
  let t = ref(but n p) in
  for i=1 to k do
    let mvts = ref(possibles !t) in
    for i=1 to random__int(list_length !mvts) do mvts := (tl !mvts) done;
    t := dplace !t (hd !mvts)
  done;
  !t
;;

(* affiche une position *)
let affiche_position t =
  let n = vect_length t.table
  and p = vect_length t.table.(0) in
  for i=0 to n-1 do
    for j=0 to p-1 do
      printf__printf " %2d" t.table.(i).(j)
    done;
    print_newline()
  done
;;

(* affiche un chemin *)
let rec affiche_chemin t mvts =
  affiche_position t;
  print_newline();
  match mvts with
  | [] -> ()
  | m::suite -> affiche_chemin (dplace t m) suite
;;

    (* +------------------------------------------------------+
       |  cherche les successeurs d'un taquin non encore vus  |
       +------------------------------------------------------+ *)

(*
   t     = position
   l     = chemin depuis la position initiale jusqu' t
   ch    = ref(liste des chemins en cours de construction)
   vus   = ref(positions dj vues)
   stop  = test d'arrt (position -> bool)
  trouv = dit si le test est positif
*)
let successeurs t l ch vus stop trouv =
  let mvts = ref(possibles t) in
  while (!mvts <> []) & (!trouv = false) do
    let m  = hd !mvts  in
    mvts := tl(!mvts);
    let t' = dplace t m in
    if not(set__mem t' !vus) then begin
      vus := set__add t' !vus;
      ch  := (t',m::l) :: !ch;
      trouv := stop(t')
    end
  done
;;


                  (* +--------------------------+
                     |  Parcours en profondeur  |
                     +--------------------------+ *)

let recherche_prof t0 t1 =
  let vus     = ref(set__add t1 (set__empty eq__compare))
  and chemins = ref[t1,[]]
  and trouv  = ref(t0 = t1)
  in
  while (!chemins <> []) & (!trouv = false) do
    let t2,l = hd(!chemins) in
    chemins := tl(!chemins);
    successeurs t2 l chemins vus (fun x -> x = t0) trouv;
  done;
  if !trouv then (rev(snd(hd !chemins)), (list_length(set__elements !vus)))
  else failwith "pas de solution"
;;

let t0 = but 3 3;;
let t1 = random_position 3 3 20;;
let sol,l = recherche_prof t0 t1;;
if list_length(sol) < 20 then affiche_chemin t1 sol;;
(*
#t1 : taquin =
 {table = [|[|1; 2; 5|]; [|3; 8; 7|]; [|6; 4; 0|]|]; i0 = 2; j0 = 2}
#sol : mouvement list =
 [Ouest; Ouest; Nord; Est; Est; Sud; Ouest; Ouest; Nord; Est; Est; Sud;
  Ouest; Ouest; Nord; Est; Est; Sud; Ouest; Ouest; Nord; Est; Est; Sud;
  Ouest; Ouest; Nord; Est; Sud; Ouest; Nord; Est; Est; Sud; Ouest; Ouest;
  Nord; Est; Est; Sud; Ouest; Ouest; Nord; Nord; Est; Est; Sud; Ouest; Ouest;
  Sud; Est; Est; Nord; Ouest; Ouest; Sud; Est; Est; Nord; Ouest; Ouest; Sud;
  Est; Est; Nord; Ouest; Ouest; Sud; Est; Est; Nord; Ouest; Ouest; Sud; Est;
  Nord; Ouest; Sud; Est; Est; Nord; Ouest; Ouest; Sud; Est; Est; Nord; Ouest;
  Ouest; Nord; Est; Est; Sud; Ouest; Ouest; Sud; Est; Est; Nord; ...]
l : int = 60960
*)

                    (* +-----------------------+
                       |  Parcours en largeur  |
                       +-----------------------+ *)


let recherche_larg t0 t1 =
  let vus     = ref(set__add t1 (set__empty eq__compare))
  and chemins = ref[t1,[]]
  and trouv  = ref(t0 = t1)
  in
  while (!chemins <> []) & (!trouv = false) do
    let ch = ref !chemins in
    chemins := [];
    while (!ch <> []) & (!trouv = false) do
      let t2,l = hd(!ch) in
      ch := tl(!ch);
      successeurs t2 l chemins vus (fun x -> x = t0) trouv
    done
  done;
  if !trouv then (rev(snd(hd !chemins)), (list_length(set__elements !vus)))
  else failwith "pas de solution"
;;

let sol,l = recherche_larg t0 t1;;
if list_length(sol) < 20 then affiche_chemin t1 sol;;
(*
#sol : mouvement list = [Nord; Ouest; Sud; Est; Nord; Nord; Ouest; Ouest]
l : int = 248
*)

          (* +------------------------------------------+
             |  Parcours en largeur par les deux bouts  |
             +------------------------------------------+ *)

let recherche_larg2 t0 t1 =
  let vus0   = ref(set__add t0 (set__empty eq__compare))
  and vus1   = ref(set__add t1 (set__empty eq__compare))
  and ch0    = ref[t0,[]]
  and ch1    = ref[t1,[]]
  and trouv = ref(t0 = t1)
  and tictac = ref(false)
  in
  while (!ch0 <> []) & (!ch1 <> []) & (!trouv = false) do
    let ch = ref !ch0 in
    ch0 := [];
    while (!ch <> []) & (!trouv = false) do
      let t2,l = hd(!ch) in
      ch := tl(!ch);
      successeurs t2 l ch0 vus0 (fun x -> set__mem x !vus1) trouv
    done;
    tictac := not(!tictac);
    let v  = !vus0 in vus0 := !vus1; vus1 := v;
    let ch = !ch0  in ch0  := !ch1;  ch1  := ch
  done;

  if !trouv then begin
    while (fst(hd !ch0)) <> (fst(hd !ch1)) do ch0 := tl(!ch0) done;
    let sol = if !tictac
    then (rev(snd(hd !ch0))) @ (map inverse (snd(hd !ch1)))
    else (rev(snd(hd !ch1))) @ (map inverse (snd(hd !ch0)))
    in (sol, list_length(set__elements !vus0) +  list_length(set__elements !vus1))
  end
  else failwith "pas de solution"
;;

let sol,l = recherche_larg2 t0 t1;;
if list_length(sol) < 20 then affiche_chemin t1 sol;;
(*
#recherche_larg2 : taquin -> taquin -> mouvement list * int = <fun>
#sol : mouvement list = [Nord; Ouest; Sud; Est; Nord; Nord; Ouest; Ouest]
l : int = 59
*)

            (* +--------------------------------------+
               |  Recherche heuristiquement ordonne  |
               +--------------------------------------+ *)

(* distance entre deux positions *)
let dist t1 t2 =
  let n   = vect_length(t1.table)
  and p   = vect_length(t1.table.(0)) in
  let pos = make_vect (n*p) (0,0)     in

  for i=0 to n-1 do for j=0 to p-1 do
    pos.(t1.table.(i).(j)) <- (i,j)
  done done;
  let d = ref 0 in
  for i=0 to n-1 do for j=0 to p-1 do
    let (x,y) = pos.(t2.table.(i).(j)) in d := !d + abs(i-x) + abs(j-y)
  done done;

  !d
;;

(* insertion dans la liste des chemins *)
let rec insre ((_,_,d) as x) ch = match ch with
  | ((_,_,d1) as x1)::suite when d1 < d -> x1::(insre x suite)
  | _ -> x::ch
;;

(* successeurs d'une position *)
let succ_heur t l ch vus t0 trouv =
  let mvts = ref(possibles t) in
  while (!mvts <> []) & (!trouv = false) do
    let m = hd(!mvts) in
    mvts := tl(!mvts);
    let t' = dplace t m in
    if not(set__mem t' !vus) then begin
      vus := set__add t' !vus;
      ch  := insre (t', m::l, dist t0 t') !ch;
      trouv := t' = t0
    end
  done
;;

let recherche_heur t0 t1 =
  let vus     = ref(set__add t1 (set__empty eq__compare))
  and chemins = ref[t1, [], dist t1 t0]
  and trouv  = ref(t0 = t1)
  in
  while (!chemins <> []) & (!trouv = false) do
    let t2,l,_ = hd(!chemins) in
    chemins := tl(!chemins);
    succ_heur t2 l chemins vus t0 trouv;
  done;
  if !trouv then let _,l,_ = hd(!chemins) in (rev(l), (list_length(set__elements !vus)))
  else failwith "pas de solution"
;;

let sol,l = recherche_heur t0 t1;;
if list_length(sol) < 20 then affiche_chemin t1 sol;;
(*
sol : mouvement list = [Nord; Ouest; Sud; Est; Nord; Nord; Ouest; Ouest]
l : int = 23
*)

(* essais avec un taquin 5*5, 30 mouvements *)
let t0 = but 5 5;;
let t1 = random_position 5 5 30;;
let sol_larg, l_larg  = recherche_larg  t0 t1;;
let sol_larg2,l_larg2 = recherche_larg2 t0 t1;;
let sol_heur, l_heur  = recherche_heur  t0 t1;;
(*
#t1 : taquin =
 {table =
   [|[|1; 6; 2; 3; 4|]; [|10; 5; 7; 8; 9|]; [|15; 11; 12; 13; 14|];
     [|20; 0; 17; 18; 19|]; [|21; 16; 22; 23; 24|]|];
  i0 = 3; j0 = 1}
#sol_larg : mouvement list = [Sud; Ouest; Nord; Nord; Nord; Est; Nord; Ouest]
l_larg : int = 2227
#sol_larg2 : mouvement list = [Sud; Ouest; Nord; Nord; Nord; Est; Nord; Ouest]
l_larg2 : int = 113
sol_heur : mouvement list =
 [Ouest; Nord; Nord; Est; Nord; Ouest; Sud; Sud; Sud; Est; Sud; Ouest; Nord;
  Nord; Nord; Nord]
l_heur : int = 75
*)
